#!/usr/bin/env python3
"""
CommandLine:
    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware='auto')
    python -m geowatch.cli.geojson_site_stats \
        --site_models="$DVC_DATA_DPATH/annotations/drop6/site_models/*" \
        --region_models="$DVC_DATA_DPATH/annotations/drop6/region_models/*"

    geowatch geomodel_stats <paths-to-site-or-region-models>
"""
import scriptconfig as scfg
import ubelt as ub


class GeojsonSiteStatsConfig(scfg.DataConfig):
    """
    Compute statistics about geojson sites.

    TODO:
        - [ ] Rename to geojson stats? Or geomodel stats?
        - [ ] make text output more  consistent and more useful.
    """
    __command__ = 'site_stats'
    __alias__ = ['geojson_stats', 'geomodel_stats']

    models = scfg.Value(None, help='site OR region models coercables (the script will attempt to distinguish them)', nargs='+', position=1)

    site_models = scfg.Value(None, help='site model coercable', nargs='+', alias=['sites'])

    region_models = scfg.Value(None, help='region model coercable', nargs='+', alias=['regions'])

    viz_dpath = scfg.Value(None, help='if specified will write stats visualizations and plots to this directory')

    io_workers = scfg.Value('avail', help='number of workers for parallel io')

    validate = scfg.Value(False, help='if True, validate the region/site models')


def main(cmdline=1, **kwargs):
    """
    Ignore:
        from geowatch.cli.geojson_site_stats import *  # NOQA
        import geowatch
        data_dvc_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        kwargs = {
            'site_models': data_dvc_dpath / 'annotations/drop6/site_models/*',
        }
        cmdline = 0

        kwargs['site_models'] = '/data/joncrall/dvc-repos/smart_expt_dvc/smartflow_evals/kit_pre_eval_8_20230131/BR_R001/sc_out_site_models'

        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware='auto')
        python -m geowatch.cli.geojson_site_stats \
            --region_models="$DVC_DATA_DPATH/annotations/drop6/region_models/PE_R001.geojson" \
            --site_models="$DVC_DATA_DPATH/annotations/drop6/site_models/PE_R001_*.geojson"
    """
    config = GeojsonSiteStatsConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    import rich
    rich.print('config = ' + ub.repr2(config))
    import copy
    import numpy as np
    import pandas as pd
    from kwutil import util_time
    from geowatch.geoannots import geomodels
    from kwgis.utils import util_gis
    from kwutil import util_parallel

    site_models = []
    region_models = []

    io_workers = util_parallel.coerce_num_workers(config['io_workers'])
    print(f'io_workers={io_workers}')

    if config.models:
        if config.site_models:
            raise ValueError('the models and site_models arguments are mutex')
        if config.region_models:
            raise ValueError('the models and region_models arguments are mutex')
        models = list(util_gis.coerce_geojson_datas(config.models, format='json', workers=io_workers))
        for model_info in models:
            model_data = model_info['data']
            model = geomodels.coerce_site_or_region_model(model_data)
            if isinstance(model, geomodels.SiteModel):
                site_models.append(model)
            elif isinstance(model, geomodels.RegionModel):
                region_models.append(model)
            else:
                raise AssertionError
    else:
        site_models = list(geomodels.SiteModel.coerce_multiple(config['site_models'], workers=io_workers))
        region_models = list(geomodels.RegionModel.coerce_multiple(config['region_models'], workers=io_workers))

    print(f'len(region_models) = {len(region_models)}')
    print(f'len(site_models) = {len(site_models)}')

    region_to_sites = ub.ddict(list)
    region_to_regions = ub.ddict(list)
    for site in site_models:
        region_to_sites[site.region_id].append(site)
    for region in region_models:
        region_to_regions[region.region_id].append(region)

    if config.validate:
        for site in site_models:
            site.validate(strict=0)
        for region in region_models:
            region.validate(strict=0)

    unique_region_ids = sorted(set(region_to_regions.keys()) | set(region_to_sites.keys()))
    print('unique_region_ids = {}'.format(ub.urepr(unique_region_ids, nl=1)))

    region_to_obs_accum = {}
    region_to_site_accum = {}

    for region_id in unique_region_ids:
        sites = region_to_sites.get(region_id, [])
        regions = region_to_regions.get(region_id, [])
        if len(regions) > 1:
            raise AssertionError('should only be 1 region file per region_id')
        elif len(regions) == 0:
            region = None
        else:
            region = regions[0]

        if region is None:
            site_summaries = []
            print(f'Region: {region_id} does not have a region model')
        else:
            site_summaries = list(region.site_summaries())
            print(f'Region: {region_id} has a region model with {len(site_summaries)} site summaries')

        print(f'Region: {region_id} has {len(sites)} site models')

        region_to_obs_accum[region_id] = region_obs_accum = []
        region_to_site_accum[region_id] = region_site_accum = []

        # Build up per-site information to expand summary stats if we have it.
        site_id_to_stats = {}
        for site in sites:
            obs_df = site.pandas_observations()
            obs_utm = util_gis.project_gdf_to_local_utm(obs_df, mode=1)
            obs_utm = geopandas_shape_stats(obs_utm)

            metric_keys = ['rt_area', 'major_obox_ratio', 'obox_major', 'obox_minor', 'hull_rt_area']
            keep_keys = ['current_phase', 'observation_date'] + metric_keys
            obs_subdf = obs_utm[keep_keys].copy()

            valid_obs_rows = obs_subdf[~pd.isnull(obs_utm['observation_date'])]
            valid_obs_rows = valid_obs_rows.sort_values('observation_date')
            obs_datetimes = [util_time.coerce_datetime(d) for d in valid_obs_rows['observation_date']]
            obs_unixtimes = [d.timestamp() for d in obs_datetimes]
            duration = np.diff(obs_unixtimes)
            obs_subdf.loc[valid_obs_rows.index[0:-1], 'duration'] = duration

            multiphase_to_duration = obs_subdf.groupby('current_phase')['duration'].sum()
            phase_to_duration = ub.ddict(lambda: 0)
            for mk, v in multiphase_to_duration.items():
                for k in mk.split(','):
                    phase_to_duration['duration.' + k.strip()] += v

            keep_keys = ['status', 'region_id', 'site_id', 'start_date', 'end_date'] + metric_keys
            site_df = site.pandas_site()
            site_df = util_gis.project_gdf_to_local_utm(site_df, mode=1)
            site_df = geopandas_shape_stats(site_df)
            site_subdf = site_df[keep_keys].copy()
            durr_df = pd.DataFrame(ub.udict(phase_to_duration).map_values(lambda x: [x]))

            site_subdf = pd.concat([site_subdf, durr_df], axis=1)

            region_obs_accum.append(obs_subdf)
            region_site_accum.append(site_subdf)

            phases = [p for p in obs_subdf['current_phase'].unique() if not pd.isnull(p)]
            site_row = copy.deepcopy(site.header['properties'])
            site_row['num_phases'] = len(phases)
            site_id_to_stats[site.site_id] = site_row
            if region is not None:
                is_earlier = region.start_date > np.array(obs_datetimes)
                is_later = region.end_date < np.array(obs_datetimes)
                num_obs_outside_time_bounds = is_earlier.sum() + is_later.sum()
                site_row['num_obs_outside_time_bounds'] = num_obs_outside_time_bounds
                site_row['num_obs'] = len(obs_utm)

        if region is not None:
            summary_df = region.pandas_summaries()
            summary_df['status'].value_counts()
            summary_df = summary_df.sort_values('status')
            summary_df = summary_df.reset_index(drop=True)
            summary_utm = util_gis.project_gdf_to_local_utm(summary_df, mode=1)

            is_sitesum_crs84_invalid = ~summary_df.geometry.is_valid
            is_sitesum_utm_invalid = ~summary_utm.geometry.is_valid
            is_sitesum_geom_invalid = (is_sitesum_crs84_invalid | is_sitesum_utm_invalid)
            summary_utm['geom_invalid'] = is_sitesum_geom_invalid

            region_df = region.pandas_region()
            region_utm = util_gis.project_gdf_to_local_utm(region_df, mode=1)

            # Find spatial intersection within the region
            if 1:
                region_start_date = region.start_date
                region_end_date = region.end_date
                gdf_site_overlaps(summary_utm, region_start_date, region_end_date)

            display_summary = summary_utm.drop(['type', 'geometry'], axis=1)
            display_summary['area_square_meters'] = summary_utm.geometry.area
            # rich.print(display_summary)

            new_rows = []
            for sitesum in display_summary.to_dict('records'):
                row = site_id_to_stats.get(sitesum['site_id'], sitesum)
                inconsistency = 0
                if row['status'] != sitesum['status']:
                    inconsistency += 1
                if row['start_date'] != sitesum['start_date']:
                    inconsistency += 1
                if row['end_date'] != sitesum['end_date']:
                    inconsistency += 1
                row.update(sitesum)
                row.pop('type', None)
                row.pop('misc_info', None)
                row.pop('cache', None)
                new_rows.append(row)

            new_sitesums = pd.DataFrame(new_rows)
            print('Site Summaries:')
            rich.print(new_sitesums.to_string())

            status_summaries = []
            for status, group in display_summary.groupby('status'):
                start_dates = [d for d in group['start_date'] if not pd.isna(d)]
                end_dates = [d for d in group['end_date'] if not pd.isna(d)]
                max_date = max(end_dates) if end_dates else None
                min_date = min(start_dates) if start_dates else None
                row = {
                    'status': status,
                    'num': len(group),
                    'min_area': group.area_square_meters.min(),
                    'max_area': group.area_square_meters.max(),
                    'min_date': min_date,
                    'max_date': max_date,
                    'invalid_geoms': group.geom_invalid.sum()
                }
                status_summaries.append(row)
            print('Summary Stats:')
            summary_stats = pd.DataFrame(status_summaries)
            rich.print(summary_stats)

            has_invalid_crs84_geom = is_sitesum_crs84_invalid.sum() > 0
            if has_invalid_crs84_geom:
                invalid_crs84_rows = summary_df[is_sitesum_crs84_invalid]
                from shapely import validation
                invalid_crs84_reasons = invalid_crs84_rows.geometry.apply(validation.explain_validity).to_list()
                print(f'invalid_crs84_reasons = {ub.urepr(invalid_crs84_reasons, nl=1)}')
                # TODO: report if we can make the polygon valid.
                try:
                    from shapely.validation import make_valid
                    invalid_crs84_rows.geometry.apply(make_valid)
                except Exception:
                    ...

            has_invalid_utm_geom = is_sitesum_utm_invalid.sum() > 0
            if has_invalid_utm_geom:
                invalid_utm_rows = summary_utm[is_sitesum_utm_invalid]
                from shapely import validation
                invalid_utm_reasons = invalid_utm_rows.geometry.apply(validation.explain_validity).to_list()
                print(f'invalid_utm_reasons = {ub.urepr(invalid_utm_reasons, nl=1)}')
                # TODO: report if we can make the polygon valid.
                try:
                    from shapely.validation import make_valid
                    invalid_utm_rows.geometry.apply(make_valid)
                except Exception:
                    ...

            import pint
            ureg = pint.UnitRegistry()
            region_area = region_utm.area.iloc[0] * (ureg.meter ** 2)
            region_area = region_area.to(ureg.kilometer ** 2)
            duration = region.end_date - region.start_date
            duration = util_time.format_timedelta(duration)

            region_stats = {
                'region_area': region_area,
                'duration': duration,
            }
            # print(f'region_stats = {ub.urepr(region_stats, nl=1)}')
            print(f'region_stats = {repr(region_stats)}')
        else:
            # Only site models are given, show their summaries
            import geopandas as gpd
            summary_rows = list(site_id_to_stats.values())
            summary_df = gpd.GeoDataFrame(summary_rows)
            rich.print(summary_df)

            if len(sites) == 1:
                # obs_df2 = pd.concat(region_obs_accum)
                # rich.print(obs_df2)
                rich.print(obs_df)
                # Show observations in this case as well
                ...

    viz_dpath = config['viz_dpath']
    if viz_dpath is not None:
        viz_dpath = ub.Path(viz_dpath).ensuredir()
        viz_site_stats(unique_region_ids, region_to_obs_accum,
                       region_to_site_accum, viz_dpath)


def gdf_site_overlaps(summary_utm, region_start_date, region_end_date):
    import kwutil
    from kwgis.utils import util_gis
    import numpy as np
    import rich
    import pandas as pd
    overlap_rows = []
    idx_to_idxs = util_gis.geopandas_pairwise_overlaps(summary_utm, summary_utm)
    for idx, idxs in idx_to_idxs.items():
        other_idxs = np.setdiff1d(idxs, [idx])
        if len(other_idxs):
            geoms1 = summary_utm.iloc[[idx]]
            geoms2 = summary_utm.iloc[other_idxs]
            s1 = geoms1.iloc[0]

            start1 = kwutil.util_time.coerce_datetime(s1['start_date']) or region_start_date
            end1 = kwutil.util_time.coerce_datetime(s1['end_date']) or region_end_date
            delta1 = end1 - start1

            g1 = s1.geometry
            isects = [g1.intersection(g2) for g2 in geoms2.geometry]
            unions = [g1.union(g2) for g2 in geoms2.geometry]
            isect_areas = np.array([g.area for g in isects])
            union_areas = np.array([g.area for g in unions])
            ious = isect_areas / union_areas
            site_id1 = geoms1['site_id'].iloc[0]
            for _ix, site_id2 in enumerate(geoms2['site_id']):
                s2 = geoms2.iloc[_ix]
                g2 = s2.geometry
                start2 = kwutil.util_time.coerce_datetime(s2['start_date']) or region_start_date
                end2 = kwutil.util_time.coerce_datetime(s2['end_date']) or region_end_date
                delta2 = end2 - start2

                start3 = max(start1, start2)
                end3 = min(end1, end2)

                start4 = min(start1, start2)
                end4 = max(end1, end2)

                isect_delta = end3 - start3
                union_delta = end4 - start4
                iot = isect_delta / union_delta

                overlap_rows.append({
                    'site_id1': site_id1,
                    'site_id2': site_id2,
                    'space_iou': ious[_ix],
                    'time_iou': iot,
                    'space_io1': isect_areas[_ix] / g1.area,
                    'space_io2': isect_areas[_ix] / g2.area,
                    'time_io1': isect_delta / delta1,
                    'time_io2': isect_delta / delta2,
                })
    overlaps = pd.DataFrame(overlap_rows)
    if len(overlaps):
        piv = overlaps.pivot(index='site_id1', columns='site_id2', values=['space_iou', 'time_iou'])
        piv = piv.fillna(0)

        # piv.sort_values('space_iou')
        # piv = piv.loc[piv.sum(axis=1).argsort().index[::-1]]
        # piv = piv[piv.sum(axis=1).sort_values().index[::-1]]
        site_order = piv['space_iou'].sum(axis=0).sort_values().index[::-1]
        piv = piv.loc[site_order]
        piv = piv.swaplevel(axis=1)[site_order]
        piv = piv[site_order]

        piv[piv == 0] = '-'
        rich.print(piv)
    else:
        rich.print('[green]no overlaps')


def viz_site_stats(unique_region_ids, region_to_obs_accum, region_to_site_accum, viz_dpath):
    from geowatch.mlops.aggregate import hash_regions
    from geowatch.utils import util_kwplot
    from geowatch.utils import util_pandas
    import matplotlib.dates as mdates
    import numpy as np
    import pandas as pd
    import rich

    import kwplot
    sns = kwplot.autosns()

    obs_stats_accum = list(ub.flatten(region_to_obs_accum.values()))
    site_stat_accum = list(ub.flatten(region_to_site_accum.values()))

    if site_stat_accum:
        site_stats = pd.concat(site_stat_accum)
    else:
        site_stats = None

    if obs_stats_accum:
        obs_stats = pd.concat(obs_stats_accum)
    else:
        obs_stats = None

    if obs_stats is not None:
        obs_stats['observation_date'] = util_kwplot.fix_matplotlib_dates(obs_stats['observation_date'])

    # dates = site_stats['start_date']
    if site_stats is not None:
        site_stats['start_date'] = util_kwplot.fix_matplotlib_dates(site_stats['start_date'])
        site_stats['end_date'] = util_kwplot.fix_matplotlib_dates(site_stats['end_date'])

        phase_duration_keys = [k for k in site_stats.columns if k.startswith('duration.')]
        for phase_duration_key in phase_duration_keys:
            deltas = site_stats[phase_duration_key]
            site_stats[phase_duration_key] = util_kwplot.fix_matplotlib_timedeltas(deltas)

    finalize_figure = util_kwplot.FigureFinalizer(
        size_inches=np.array([6.4, 4.8]) * 1.0,
        dpath=viz_dpath,
    )

    fnum = 1

    def new_figure():
        fig = kwplot.figure(doclf=True, fnum=fnum)
        # if 0:
        #     fnum += 1
        return fig

    kwplot.close_figures()

    def corr_label(group, x, y):
        metrics_of_interest = group[[x, y]]
        metric_corr_mat = metrics_of_interest.corr()
        metric_corr = metric_corr_mat.stack()
        metric_corr.name = 'corr'
        stack_idx = metric_corr.index
        valid_idxs = [(a, b) for (a, b) in ub.unique(map(tuple, map(sorted, stack_idx.to_list()))) if a != b]
        if valid_idxs:
            metric_corr = metric_corr.loc[valid_idxs]
            # corr_lbl = 'corr({},{})={:0.4f}'.format(*metric_corr.index[0], metric_corr.iloc[0])
            corr_lbl = 'corr={:0.4f}'.format(metric_corr.iloc[0])
        else:
            corr_lbl = ''
        return corr_lbl

    ### PER-SITE PLOTS
    if site_stats is not None:
        label_modifier = util_kwplot.LabelModifier()
        for phase_duration_key in phase_duration_keys:
            x = phase_duration_key
            label_modifier.add_mapping({x: x + ' (days)'})
            label_modifier.add_mapping({'rt_area': 'sqrt(area)'})

        regions = site_stats['region_id'].unique().tolist()
        region_title = hash_regions(regions)

        ax = new_figure().gca()
        sns.boxplot(data=site_stats, x='status', y='end_date', ax=ax)
        ax.set_title(f'End Date Distribution: regions={region_title}')
        ax.yaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d'))
        ax.yaxis.set_major_locator(mdates.DayLocator(interval=356))
        finalize_figure(ax.figure, 'date_distri_stop.png')

        ax = new_figure().gca()
        sns.boxplot(data=site_stats, x='status', y='start_date', ax=ax)
        ax.set_title(f'Start Date Distribution: regions={region_title}')
        ax.yaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d'))
        ax.yaxis.set_major_locator(mdates.DayLocator(interval=356))
        finalize_figure(ax.figure, 'date_distri_start.png')

        ax = new_figure().gca()
        ax.set_title(f'regions={region_title}')
        sns.boxplot(data=site_stats, x='status', y='rt_area', ax=ax)
        finalize_figure(ax.figure, 'status_vs_area.png')

        ax = new_figure().gca()
        sns.scatterplot(data=site_stats, x='obox_major', y='obox_minor', ax=ax, hue='status')
        ax.set_title(f'regions={region_title}')
        finalize_figure(ax.figure, 'obox_aspect_ratio.png')

        ax = new_figure().gca()
        sns.scatterplot(data=site_stats, x='obox_major', y='obox_minor', ax=ax, hue='status')
        ax.set_title(f'regions={region_title}')
        finalize_figure(ax.figure, 'obox_aspect_ratio.png')

        phase_duration_keys = [k for k in site_stats.columns if k.startswith('duration.')]
        max_durration = site_stats[phase_duration_keys].max().max()

        table = site_stats[phase_duration_keys].describe().applymap(lambda x: ''.join(str(x).partition('days')[0:2]))
        table = util_pandas.pandas_shorten_columns(table)
        rich.print(table)

        y = 'rt_area'
        ddd = site_stats[phase_duration_keys + ['site_id', 'rt_area']].melt(id_vars=['site_id', 'rt_area'])
        ddd['value'] = [getattr(d, 'days', None) for d in ddd['value']]
        ax = new_figure().gca()
        sns.scatterplot(data=ddd, x='value', y=y, ax=ax, hue='variable')
        ax.set_title(f'Durations\nregions={region_title}')
        ax.set_xscale('log')
        ax.set_yscale('log')
        for k, group in ddd.groupby('variable'):
            x = 'value'
            label_modifier.update({k: k + ': ' + corr_label(group, x, y)})
        label_modifier(ax)
        finalize_figure(ax.figure, 'area_vs_duration_melt.png')

        for k, group in ddd.groupby('variable'):
            label_modifier.update({k: k})
        ax = new_figure().gca()
        sns.boxplot(data=ddd, x='variable', y=x, ax=ax)
        ax.set_ylim(1, max_durration.days)
        label_modifier(ax)
        ax.set_title(f'Durations\nregions={region_title}')
        finalize_figure(ax.figure, 'duration_boxplot_melt.png')

        for phase_duration_key in phase_duration_keys:
            ax = new_figure().gca()
            valid_duration_df = site_stats[~pd.isnull(site_stats[phase_duration_key])].copy()
            valid_duration_df[phase_duration_key] = [d.days for d in valid_duration_df[phase_duration_key]]
            hue = 'status'
            x = phase_duration_key
            y = 'rt_area'
            for k, group in valid_duration_df.groupby(hue):
                label_modifier.update({k: k + ': ' + corr_label(group, x, y)})
                ...
            sns.scatterplot(data=valid_duration_df, x=x, y=y, ax=ax, hue=hue)
            ax.set_title(f'{phase_duration_key}\nregions={region_title}')
            ax.set_xscale('log')
            ax.set_yscale('log')
            label_modifier(ax)
            finalize_figure(ax.figure, f'area_vs_{phase_duration_key}.png')

            for k, group in valid_duration_df.groupby(hue):
                label_modifier.update({k: k})
            ax = new_figure().gca()
            sns.boxplot(data=valid_duration_df, x=hue, y=x, ax=ax)
            ax.set_ylim(1, max_durration.days)
            label_modifier(ax)
            ax.set_title(f'{phase_duration_key}\nregions={region_title}')
            finalize_figure(ax.figure, f'duration_boxplot_{phase_duration_key}.png')

        # with sns.color_palette("flare"):
        # sns.scatterplot(data=site_stats, x='rt_area', y='major_obox_ratio', hue='status')

    ### PER-OBSERVATION PLOTS
    if obs_stats is not None:
        # regions = obs_stats['region_id'].unique().tolist()
        # region_title = hash_regions(regions)
        region_title = hash_regions(unique_region_ids)

        ax = new_figure().gca()
        ax.set_title(f'regions={regions}')
        sns.boxplot(data=obs_stats, x='current_phase', y='observation_date', ax=ax)
        ax.yaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d'))
        ax.yaxis.set_major_locator(mdates.DayLocator(interval=356))
        finalize_figure(ax.figure, 'phase_vs_obsdate.png')

        ax = kwplot.figure(doclf=True, fnum=4).gca()
        ax.set_title(f'regions={regions}')
        sns.boxplot(data=obs_stats, x='current_phase', y='rt_area', ax=ax)
        finalize_figure(ax.figure, 'phase_vs_size.png')

        ax = kwplot.figure(doclf=True, fnum=5).gca()
        sns.scatterplot(data=obs_stats, x='obox_major', y='obox_minor', ax=ax, hue='current_phase')
        ax.set_title(f'regions={regions}')
        finalize_figure(ax.figure, 'obox_aspect_ratio_phase.png')

        ax = kwplot.figure(doclf=True, fnum=5).gca()
        sns.scatterplot(data=obs_stats, x='obox_major', y='obox_minor', ax=ax, hue='current_phase')
        ax.set_title(f'regions={regions}')
        finalize_figure(ax.figure, 'size_verus_duration.png')

    # import xdev
    # xdev.view_diri


def geopandas_shape_stats(df):
    """
    Compute shape statistics about a geopandas dataframe (assume UTM CRS)
    """
    import kwimage
    import numpy as np
    # df['area'] = df.geometry.area
    # df['hull_area'] = df.geometry.convex_hull.area
    df['hull_rt_area'] = np.sqrt(df.geometry.convex_hull.area)
    df['rt_area'] = np.sqrt(df.geometry.area)

    obox_whs = [kwimage.MultiPolygon.from_shapely(s).oriented_bounding_box().extent
                for s in df.geometry]

    df['obox_major'] = [max(e) for e in obox_whs]
    df['obox_minor'] = [min(e) for e in obox_whs]
    df['major_obox_ratio'] = df['obox_major'] / df['obox_minor']

    # df['ch_aspect_ratio'] =
    # df['isoperimetric_quotient'] = df.geometry.apply(shapestats.ipq)
    # df['boundary_amplitude'] = df.geometry.apply(shapestats.compactness.boundary_amplitude)
    # df['eig_seitzinger'] = df.geometry.apply(shapestats.compactness.eig_seitzinger)
    return df


__cli__ = GeojsonSiteStatsConfig
__cli__.main = main

if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/geowatch/cli/geojson_site_stats.py
    """
    main()



#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub


class DiskInfoCLI(scfg.DataConfig):
    """
    Get information about the underlying storage medium used by a path.
    """
    dpath = scfg.Value('.', help='The path to inspect')
    key = scfg.Value(None, help=ub.paragraph(
        '''
        If specified print only the value of this key of interest, otherwise
        print all info. Useful keys are "hwtype", "filesystem".
        '''))
    verbose = scfg.Value(0, help='verbosity level')

    @classmethod
    def main(cls, cmdline=1, **kwargs):
        """
        Example:
            >>> # xdoctest: +SKIP
            >>> from geowatch.cli.experimental.disk_info import *  # NOQA
            >>> cmdline = 0
            >>> kwargs = dict()
            >>> cls = DiskInfoCLI
            >>> cls.main(cmdline=cmdline, **kwargs)
        """
        import rich
        config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
        if config.verbose:
            rich.print('config = ' + ub.urepr(config, nl=1))

        from geowatch.utils.util_hardware import disk_info_of_path
        info = disk_info_of_path('.')

        if config.key is None:
            rich.print(f'info = {ub.urepr(info, nl=1)}')
        else:
            print(info[config.key])

__cli__ = DiskInfoCLI
main = __cli__.main

if __name__ == '__main__':
    """

    CommandLine:
        python ~/code/geowatch/geowatch/cli/experimental/disk_info.py
        python -m geowatch.cli.experimental.disk_info --key hwtype
    """
    main()



"""
Moved to:
    ~/code/kwcoco_dataloader/kwcoco_dataloader/tasks/fusion/datamodules/temporal_sampling/exceptions.py
"""
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import TimeSampleError  # NOQA



"""
Originally these were doctests in the dataloader, but were removed for space.

TODO: These should be refactored into more descriptive tests that target
specific parameters in the datamodule config
"""
from geowatch.tasks.fusion.datamodules.kwcoco_dataset import KWCocoVideoDataset
import ndsampler
import kwcoco
import geowatch


def test_watch_tasks_fusion_datamodules_kwcoco_dataset_full_KWCocoVideoDataset():
    """
    converted from /home/joncrall/code/watch/geowatch/tasks/fusion/datamodules/_orig_kwcoco_dataset_full.py::KWCocoVideoDataset:1
    """
    coco_dset = geowatch.coerce_kwcoco('geowatch')
    print({c.get('sensor_coarse') for c in coco_dset.images().coco_images})
    print({c.channels.spec for c in coco_dset.images().coco_images})
    sampler = ndsampler.CocoSampler(coco_dset)
    self = KWCocoVideoDataset(sampler, time_dims=2, window_dims=(128, 128),
                              channels=None)
    index = 0
    item = self[index]
    canvas = self.draw_item(item)
    canvas


def test_watch_tasks_fusion_datamodules_kwcoco_dataset_full_KWCocoVideoDataset___getitem__1():
    """
    converted from /home/joncrall/code/watch/geowatch/tasks/fusion/datamodules/_orig_kwcoco_dataset_full.py::KWCocoVideoDataset.__getitem__:1
    """
    coco_dset = kwcoco.CocoDataset.demo('vidshapes2-multispectral', num_frames=5)
    sampler = ndsampler.CocoSampler(coco_dset)
    channels = 'B10|B8a|B1|B8'
    self = KWCocoVideoDataset(sampler, time_dims=5, window_dims=(530, 610),
                              channels=channels, dist_weights=1,
                              temporal_dropout=0.5)
    item = self[0]
    canvas = self.draw_item(item)
    canvas


def test_watch_tasks_fusion_datamodules_kwcoco_dataset_full_KWCocoVideoDataset___getitem__2():
    """
    converted from /home/joncrall/code/watch/geowatch/tasks/fusion/datamodules/_orig_kwcoco_dataset_full.py::KWCocoVideoDataset.__getitem__:2
    """
    coco_dset = kwcoco.CocoDataset.demo('vidshapes2-multispectral', num_frames=5)
    sampler = ndsampler.CocoSampler(coco_dset)
    channels = 'B10|B8|B1'
    self = KWCocoVideoDataset(sampler, time_dims=4, window_dims=(96, 96),
                              channels=channels, neg_to_pos_ratio=0.1)
    item = self[-1]
    canvas = self.draw_item(item)  # NOQA


def test_oob_target():
    import ubelt as ub
    coco_dset = geowatch.coerce_kwcoco('vidshapes8')
    sampler = ndsampler.CocoSampler(coco_dset)
    self = KWCocoVideoDataset(sampler, time_dims=1, window_dims=(128, 128),
                              channels=None)
    target = {
        'main_idx': 0,
        'gids': [2],
        'space_slice':  (
            # slice(-155.0, 357.0, None),
            # slice(-256.0, 256.0, None)
            # slice(20.0, 350, None),
            # slice(20.0, 130, None)
            slice(-200.0, 350, None),
            slice(-200.0, 350, None)
        ),
        'allow_augment': False,
    }
    index = target
    item = self[index]
    summary = self.summarize_item(item)
    print(f'summary = {ub.urepr(summary, nl=-1)}')

    s1 = h1, w1, c1 = item['frames'][0]['class_ohe'].shape
    s2 = c2, h2, w2 = item['frames'][0]['modes']['r|g|b'].shape
    s3 = h3, w3 = item['frames'][0]['saliency'].shape
    print(s1)
    print(s2)
    print(s3)
    assert h1 == h2
    assert w1 == w2

    if 0:
        import numpy as np
        import kwplot
        kwplot.autompl()
        canvas = self.draw_item(item)  # NOQA
        kwplot.imshow(canvas, pnum=(1, 3, 1), fnum=1)

        canvas = self.sampler.dset.coco_image(target['gids'][0]).imdelay().finalize()
        kwplot.imshow(canvas, pnum=(1, 3, 2), fnum=1)

        canvas = item['frames'][0]['modes']['r|g|b'].numpy().transpose(1, 2, 0).astype(np.uint8)
        kwplot.imshow(canvas, pnum=(1, 3, 3), fnum=1)


def test_resolution_on_nongeo_dataset():
    """
    The resolution parameter should help determine a scale factor on
    non-geo datasets
    """
    import ubelt as ub
    coco_dset = geowatch.coerce_kwcoco('vidshapes8')
    sampler = ndsampler.CocoSampler(coco_dset)
    self = KWCocoVideoDataset(
        sampler,
        time_dims=1,
        window_dims=(128, 128),
        channels=None,
        window_resolution=1,
        input_resolution=0.5,
        output_resolution=0.5,
        use_grid_positives=1,
        use_centered_positives=0,
    )
    index = 0
    item = self[index]
    target = item['target']
    summary = self.summarize_item(item)
    print(f'summary = {ub.urepr(summary, nl=-1)}')

    s1 = h1, w1, c1 = item['frames'][0]['class_ohe'].shape
    s2 = c2, h2, w2 = item['frames'][0]['modes']['r|g|b'].shape
    s3 = h3, w3 = item['frames'][0]['saliency'].shape
    print(s1)
    print(s2)
    print(s3)
    assert h1 == h2
    assert w1 == w2

    if 1:
        import numpy as np
        import kwplot
        kwplot.autompl()
        canvas = self.draw_item(item)  # NOQA
        kwplot.imshow(canvas, pnum=(1, 3, 1), fnum=1)

        canvas = self.sampler.dset.coco_image(target['gids'][0]).imdelay().finalize()
        kwplot.imshow(canvas, pnum=(1, 3, 2), fnum=1)

        canvas = item['frames'][0]['modes']['r|g|b'].numpy().transpose(1, 2, 0).astype(np.uint8)
        kwplot.imshow(canvas, pnum=(1, 3, 3), fnum=1)


def test_nonlocal_perframe_classification_labels():
    coco_dset = geowatch.coerce_kwcoco('vidshapes8')
    sampler = ndsampler.CocoSampler(coco_dset)
    self = KWCocoVideoDataset(
        sampler,
        time_dims=1,
        window_dims=(128, 128),
        channels=None,
        window_resolution=1,
        input_resolution=0.5,
        output_resolution=0.5,
        use_grid_positives=1,
        use_centered_positives=0,
    )
    self.requested_tasks['nonlocal_class'] = True
    self.disable_augmenter = True
    index = 0
    item = self[index]
    # ensure the nonlocal class one-hot-embedding exists
    nonlocal_class_ohe = item['frames'][0]['nonlocal_class_ohe']
    import torch
    assert torch.is_tensor(nonlocal_class_ohe)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: deeplab2/evaluator.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='deeplab2/evaluator.proto',
    package='deeplab2',
    syntax='proto2',
    serialized_options=None,
    serialized_pb=b'\n\x18\x64\x65\x65plab2/evaluator.proto\x12\x08\x64\x65\x65plab2\"\xe5\x05\n\x10\x45valuatorOptions\x12\x16\n\neval_steps\x18\x01 \x01(\x05:\x02-1\x12\x1b\n\reval_interval\x18\x02 \x01(\x05:\x04\x35\x30\x30\x30\x12#\n\x17\x63ontinuous_eval_timeout\x18\x03 \x01(\x05:\x02-1\x12\x1d\n\x0fuse_tf_function\x18\x04 \x01(\x08:\x04true\x12\x1b\n\x10stuff_area_limit\x18\x06 \x01(\x05:\x01\x30\x12\x1b\n\x10thing_area_limit\x18\x13 \x01(\x05:\x01\x30\x12\x33\n&transformer_class_confidence_threshold\x18\x14 \x01(\x02:\x03\x30.7\x12\'\n\x1apixel_confidence_threshold\x18\x15 \x01(\x02:\x03\x30.4\x12#\n\x16\x63\x65nter_score_threshold\x18\x07 \x01(\x02:\x03\x30.1\x12\x15\n\nnms_kernel\x18\x08 \x01(\x05:\x01\x33\x12\x1b\n\x0ekeep_k_centers\x18\t \x01(\x05:\x03\x34\x30\x30\x12\x1f\n\x10save_predictions\x18\n \x01(\x08:\x05\x66\x61lse\x12\x19\n\x11override_save_dir\x18\x0b \x01(\t\x12\x1b\n\x0fnum_vis_samples\x18\x0c \x01(\x05:\x02\x31\x30\x12#\n\x14save_raw_predictions\x18\r \x01(\x08:\x05\x66\x61lse\x12,\n\x13raw_panoptic_format\x18\x11 \x01(\t:\x0ftwo_channel_png\x12%\n\x17\x63onvert_raw_to_eval_ids\x18\x0e \x01(\x08:\x04true\x12!\n\x12\x61\x64\x64_flipped_images\x18\x05 \x01(\x08:\x05\x66\x61lse\x12\x17\n\x0b\x65val_scales\x18\x0f \x03(\x02\x42\x02\x10\x01\x12\x34\n&merge_semantic_and_instance_with_tf_op\x18\x10 \x01(\x08:\x04true\x12\"\n\x13\x64\x65tailed_ap_metrics\x18\x12 \x01(\x08:\x05\x66\x61lse'
)

_EVALUATOROPTIONS = _descriptor.Descriptor(
    name='EvaluatorOptions',
    full_name='deeplab2.EvaluatorOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='eval_steps', full_name='deeplab2.EvaluatorOptions.eval_steps', index=0,
            number=1, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=-1,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='eval_interval', full_name='deeplab2.EvaluatorOptions.eval_interval', index=1,
            number=2, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=5000,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='continuous_eval_timeout', full_name='deeplab2.EvaluatorOptions.continuous_eval_timeout', index=2,
            number=3, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=-1,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='use_tf_function', full_name='deeplab2.EvaluatorOptions.use_tf_function', index=3,
            number=4, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=True,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='stuff_area_limit', full_name='deeplab2.EvaluatorOptions.stuff_area_limit', index=4,
            number=6, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='thing_area_limit', full_name='deeplab2.EvaluatorOptions.thing_area_limit', index=5,
            number=19, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='transformer_class_confidence_threshold',
            full_name='deeplab2.EvaluatorOptions.transformer_class_confidence_threshold', index=6,
            number=20, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0.7),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='pixel_confidence_threshold', full_name='deeplab2.EvaluatorOptions.pixel_confidence_threshold',
            index=7,
            number=21, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0.4),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='center_score_threshold', full_name='deeplab2.EvaluatorOptions.center_score_threshold', index=8,
            number=7, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0.1),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='nms_kernel', full_name='deeplab2.EvaluatorOptions.nms_kernel', index=9,
            number=8, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=3,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='keep_k_centers', full_name='deeplab2.EvaluatorOptions.keep_k_centers', index=10,
            number=9, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=400,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='save_predictions', full_name='deeplab2.EvaluatorOptions.save_predictions', index=11,
            number=10, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='override_save_dir', full_name='deeplab2.EvaluatorOptions.override_save_dir', index=12,
            number=11, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='num_vis_samples', full_name='deeplab2.EvaluatorOptions.num_vis_samples', index=13,
            number=12, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=10,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='save_raw_predictions', full_name='deeplab2.EvaluatorOptions.save_raw_predictions', index=14,
            number=13, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='raw_panoptic_format', full_name='deeplab2.EvaluatorOptions.raw_panoptic_format', index=15,
            number=17, type=9, cpp_type=9, label=1,
            has_default_value=True, default_value=b"two_channel_png".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='convert_raw_to_eval_ids', full_name='deeplab2.EvaluatorOptions.convert_raw_to_eval_ids', index=16,
            number=14, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=True,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='add_flipped_images', full_name='deeplab2.EvaluatorOptions.add_flipped_images', index=17,
            number=5, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='eval_scales', full_name='deeplab2.EvaluatorOptions.eval_scales', index=18,
            number=15, type=2, cpp_type=6, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=b'\020\001', file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='merge_semantic_and_instance_with_tf_op',
            full_name='deeplab2.EvaluatorOptions.merge_semantic_and_instance_with_tf_op', index=19,
            number=16, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=True,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='detailed_ap_metrics', full_name='deeplab2.EvaluatorOptions.detailed_ap_metrics', index=20,
            number=18, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=39,
    serialized_end=780,
)

DESCRIPTOR.message_types_by_name['EvaluatorOptions'] = _EVALUATOROPTIONS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EvaluatorOptions = _reflection.GeneratedProtocolMessageType('EvaluatorOptions', (_message.Message,), {
    'DESCRIPTOR': _EVALUATOROPTIONS,
    '__module__': 'deeplab2.evaluator_pb2'
    # @@protoc_insertion_point(class_scope:deeplab2.EvaluatorOptions)
})
_sym_db.RegisterMessage(EvaluatorOptions)

_EVALUATOROPTIONS.fields_by_name['eval_scales']._options = None
# @@protoc_insertion_point(module_scope)



"""
This describes the structure of how a task prediction script should work.


What must happen is that there must be a predict.py script that accepts a
kwcoco as input and produces a kwcoco as output. The output kwcoco should have
additional "auxiliary" or "asset" images added to each "coco image". These new
assets are rasters that can contain any number of channels.

These are the features that will be used as input to the fusion network.

The existing invariants/predict.py code (as well as fusion/predict.py and every
other task/*/predict.py) has logic to run a model over a sliding window and
stitch the rasters back together such that they align to the larger image they
belong to (up to an affine transformation).


For example an input coco image dictionary may look like this:

{
    'date_captured': '1984-05-24T14:44:38.034876',
    'file_name': None,
    'frame_index': 0,
    'height': 661,
    'width': 778
    'id': 1,
    'name': 'generated-1-0',
    'sensor': 'sensor1',
    'sensor_coarse': 'sensor0',
    'valid_region': [{
        'exterior':
            [[0.0000, 0.0000], [0.0000, 661.0000], [778.0000, 661.0000],
            [778.0000, 0.0000], [0.0000, 0.0000]
        ],
        'interiors': []}],
    'valid_region_utm': {
        'coordinates': [
        [[[496408.5193, -5231846.5362], [496408.5193, -5231712.3362], [496530.5193, -5231712.3362],
        [496530.5193, -5231846.5362], [496408.5193, -5231846.5362]]]],
        'properties': {
            'crs': {'auth': ['EPSG', '32620'], 'axis_mapping': 'OAMS_AUTHORITY_COMPLIANT'}
        },
        'type': 'MultiPolygon'
    },
    'geos_corners': {
        'coordinates': [
            [[-63.0475, -47.2401], [-63.0475, -47.2389],
                [-63.0458, -47.2389], [-63.0458, -47.2401]] ],
        'properties': {
            'crs_info': {'auth': ['EPSG', '4326'], 'axis_mapping': 'OAMS_TRADITIONAL_GIS_ORDER'}
        },
        'type': 'Polygon'
    },
    'video_id': 1,
    'warp_img_to_vid': {
        'offset': [0.0000, 0.0000],
        'scale': [0.7236, 0.5203],
        'shearx': -0.0000,
        'theta': -0.0000,
        'type': 'affine'
    },

    'auxiliary': [
        {
            'approx_meter_gsd': 0.1792,
            'band_metas': [{'nodata': None}],
            'height': 661,
            'width': 778,
            'num_bands': 1,
            'is_rpc': False,
            'channels': 'B1',
            'file_name': '_assets/aux/aux_B1/img_00001.tif',
            'warp_aux_to_img': {'type': 'affine', 'scale': 1},
        }

        {
            'approx_meter_gsd': 0.5638,
            'band_metas': [{'nodata': None}],
            'channels': 'B8',
            'file_name': '_assets/aux/aux_B8/img_00001.tif',
            'height': 111,
            'is_rpc': False,
            'num_bands': 1,
            'warp_aux_to_img': {'scale': 6.0000, 'type': 'affine'},
            'width': 130,
        },

    ]
}

"""


def demo():
    # This demo will show how to add new assets / auxiliary images to a kwcoco
    # file.
    import watch
    import kwcoco
    import kwimage
    import numpy as np
    import ubelt as ub

    # Create a path where we can dumpy demo info
    demo_dpath = ub.Path.appdir('watch/demo/demo_add_auxiliary').ensuredir()

    # Start off by loading a demo dataset (this would be your dataset)
    dset = watch.coerce_kwcoco('watch-msi', geodata=True, dates=True)

    # We are going to add a new asset to each image.
    for image_id in dset.images():
        # Create a CocoImage object for each image.
        coco_image: kwcoco.CocoImage = dset.coco_image(image_id)

        image_name = coco_image.img['name']

        # When we add an asset we need to pass it the file path to that asset
        # and some additional information, namely the:
        # 1. filepath of the new asset
        # 2. the transform that warps the asset into "image space"
        # 3. Information about what channels / bands are in the image.
        # Other information can be added, but this tutorial currently only
        # covers the simple case.
        #coco_image.add_asset()

        # Let's pretend we've made a new asset for this coco image with 4 new
        # bands I will label as "foo" "bar" "baz" and "biz".

        # Let's create the demo image
        img_w = coco_image.img['width']
        img_h = coco_image.img['height']

        asset_w = img_w
        asset_h = img_h
        imdata = np.random.rand(asset_h, asset_w , 4)
        new_fpath = demo_dpath / f'image_{image_name}_myfeat.tif'
        kwimage.imwrite(new_fpath, imdata)

        # We will label it with this channel code:
        channels = 'foo|bar|baz|biz'

        # We need the transform that warps from asset space to image space
        # In this case they are aligned so we can just use the identity.
        warp_aux_to_img = kwimage.Affine.eye()

        # Use the CocoImage helper which will augment the coco dictionary with
        # your information.
        coco_image.add_asset(new_fpath, channels=channels, width=asset_w,
                             height=asset_h, warp_aux_to_img=warp_aux_to_img)



"""
Note:
    This module has been moved to :mod:`kwgis.gis.geotiff`.
"""
from kwgis.gis.geotiff import geotiff_metadata, geotiff_crs_info, make_crs_info_object, geotiff_filepath_info, parse_landsat_product_id, walk_geotiff_products  # NOQA



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the code for the Motion-DeepLab decoder."""

import tensorflow as tf

from deeplab2 import common
from deeplab2 import config_pb2
from deeplab2.model.decoder import panoptic_deeplab


class MotionDeepLabDecoder(tf.keras.layers.Layer):
    """A Motion-DeepLab decoder layer.

    This layer takes low- and high-level features as input and uses a dual-ASPP
    and dual-decoder structure to aggregate features for semantic and instance
    segmentation. On top of the decoders, four heads are used to predict semantic
    segmentation, instance center probabilities, instance center regression, and
    previous frame offset regression per pixel.
    """

    def __init__(
            self,
            decoder_options: config_pb2.DecoderOptions,
            motion_deeplab_options: config_pb2.ModelOptions.MotionDeepLabOptions,
            bn_layer=tf.keras.layers.BatchNormalization):
        """Initializes a Motion-DeepLab decoder.

        Args:
          decoder_options: Decoder options as defined in config_pb2.DecoderOptions.
          motion_deeplab_options: Model options as defined in
            config_pb2.ModelOptions.MotionDeeplabOptions.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
        """
        super(MotionDeepLabDecoder, self).__init__(name='MotionDeepLabDecoder')

        low_level_feature_keys = [
            item.feature_key for item in motion_deeplab_options.low_level
        ]
        low_level_channels_project = [
            item.channels_project for item in motion_deeplab_options.low_level
        ]

        self._semantic_decoder = panoptic_deeplab.PanopticDeepLabSingleDecoder(
            decoder_options.feature_key,
            low_level_feature_keys,
            low_level_channels_project,
            decoder_options.aspp_channels,
            decoder_options.decoder_channels,
            decoder_options.atrous_rates,
            name='semantic_decoder',
            bn_layer=bn_layer)
        self._semantic_head = panoptic_deeplab.PanopticDeepLabSingleHead(
            motion_deeplab_options.semantic_head.head_channels,
            motion_deeplab_options.semantic_head.output_channels,
            common.PRED_SEMANTIC_LOGITS_KEY,
            name='semantic_head',
            bn_layer=bn_layer)

        self._instance_decoder = None
        self._instance_center_head = None
        self._instance_regression_head = None
        self._motion_regression_head = None

        if motion_deeplab_options.instance.low_level_override:
            low_level_options = motion_deeplab_options.instance.low_level_override
        else:
            low_level_options = motion_deeplab_options.low_level

        # If instance_decoder is set, use those options; otherwise reuse the
        # architecture as defined for the semantic decoder.
        if motion_deeplab_options.instance.HasField('instance_decoder_override'):
            decoder_options = (motion_deeplab_options.instance
                               .instance_decoder_override)

        low_level_feature_keys = [item.feature_key for item in low_level_options]
        low_level_channels_project = [
            item.channels_project for item in low_level_options
        ]

        self._instance_decoder = panoptic_deeplab.PanopticDeepLabSingleDecoder(
            decoder_options.feature_key,
            low_level_feature_keys,
            low_level_channels_project,
            decoder_options.aspp_channels,
            decoder_options.decoder_channels,
            decoder_options.atrous_rates,
            name='instance_decoder',
            bn_layer=bn_layer)
        self._instance_center_head = panoptic_deeplab.PanopticDeepLabSingleHead(
            motion_deeplab_options.instance.center_head.head_channels,
            motion_deeplab_options.instance.center_head.output_channels,
            common.PRED_CENTER_HEATMAP_KEY,
            name='instance_center_head',
            bn_layer=bn_layer)
        self._instance_regression_head = panoptic_deeplab.PanopticDeepLabSingleHead(
            motion_deeplab_options.instance.regression_head.head_channels,
            motion_deeplab_options.instance.regression_head.output_channels,
            common.PRED_OFFSET_MAP_KEY,
            name='instance_regression_head',
            bn_layer=bn_layer)

        # The motion head regresses every pixel to its center in the previous
        # frame.
        self._motion_regression_head = panoptic_deeplab.PanopticDeepLabSingleHead(
            motion_deeplab_options.motion_head.head_channels,
            motion_deeplab_options.motion_head.output_channels,
            common.PRED_FRAME_OFFSET_MAP_KEY,
            name='motion_regression_head',
            bn_layer=bn_layer)

    def reset_pooling_layer(self):
        """Resets the ASPP pooling layers to global average pooling."""
        self._semantic_decoder.reset_pooling_layer()
        if self._instance_decoder is not None:
            self._instance_decoder.reset_pooling_layer()

    def set_pool_size(self, pool_size):
        """Sets the pooling size of the ASPP pooling layers.

        Args:
          pool_size: A tuple specifying the pooling size of the ASPP pooling layers.
        """
        self._semantic_decoder.set_pool_size(pool_size)
        if self._instance_decoder is not None:
            self._instance_decoder.set_pool_size(pool_size)

    def get_pool_size(self):
        return self._semantic_decoder.get_pool_size()

    def call(self, features, training=False):
        """Performs a forward pass.

        Args:
          features: An input dict of tf.Tensor with shape [batch, height, width,
            channels]. Different keys should point to different features extracted
            by the encoder, e.g. low-level or high-level features.
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          A dictionary containing the results of the semantic segmentation head and
            depending on the configuration also of the instance segmentation head.
        """

        semantic_features = self._semantic_decoder(features, training=training)
        results = self._semantic_head(semantic_features, training=training)

        if self._instance_decoder is not None:
            instance_features = self._instance_decoder(features, training=training)
            instance_center_predictions = self._instance_center_head(
                instance_features, training=training)
            instance_regression_predictions = self._instance_regression_head(
                instance_features, training=training)
            motion_regression_predictions = self._motion_regression_head(
                instance_features, training=training)
            if results.keys() & motion_regression_predictions.keys():
                raise ValueError('The keys of the semantic branch and the instance '
                                 'motion branch overlap. Please use unique keys.')
            results.update(motion_regression_predictions)

            if results.keys() & instance_center_predictions.keys():
                raise ValueError('The keys of the semantic branch and the instance '
                                 'center branch overlap. Please use unique keys.')
            results.update(instance_center_predictions)

            if results.keys() & instance_regression_predictions.keys():
                raise ValueError('The keys of the semantic branch and the instance '
                                 'regression branch overlap. Please use unique keys.')
            results.update(instance_regression_predictions)

        return results

    @property
    def checkpoint_items(self):
        items = {
            common.CKPT_SEMANTIC_DECODER:
                self._semantic_decoder,
            common.CKPT_SEMANTIC_HEAD_WITHOUT_LAST_LAYER:
                self._semantic_head.conv_block,
            common.CKPT_SEMANTIC_LAST_LAYER:
                self._semantic_head.final_conv
        }
        if self._instance_decoder is not None:
            instance_items = {
                common.CKPT_INSTANCE_DECODER:
                    self._instance_decoder,
                common.CKPT_INSTANCE_CENTER_HEAD_WITHOUT_LAST_LAYER:
                    self._instance_center_head.conv_block,
                common.CKPT_INSTANCE_CENTER_HEAD_LAST_LAYER:
                    self._instance_center_head.final_conv,
                common.CKPT_INSTANCE_REGRESSION_HEAD_WITHOUT_LAST_LAYER:
                    self._instance_regression_head.conv_block,
                common.CKPT_INSTANCE_REGRESSION_HEAD_LAST_LAYER:
                    self._instance_regression_head.final_conv,
                common.CKPT_MOTION_REGRESSION_HEAD_WITHOUT_LAST_LAYER:
                    self._motion_regression_head.conv_block,
                common.CKPT_MOTION_REGRESSION_HEAD_LAST_LAYER:
                    self._motion_regression_head.final_conv,
            }
            items.update(instance_items)
        return items



"""
aggregation functions for heatmaps
"""


def _norm(heatmaps, norm_ord):
    """
    Computes the generalized mean over axis=0.

    Args:
        heatmaps (List[ndarray]) pixel aligned heatmaps
        norm_ord (int | float): the exponent of the generalized mean.

    Returns:
        ndarray : the axis=0 is marginalized over.

    Notes:
        like np.linalg.norm but with special nan handling and a division factor

    References:
        https://en.wikipedia.org/wiki/Generalized_mean
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.pmean.html

    Example:
        >>> from geowatch.tasks.tracking.agg_functions import *  # NOQA
        >>> from geowatch.tasks.tracking.agg_functions import _norm
        >>> import ubelt as ub
        >>> import kwimage
        >>> import numpy as np
        >>> num_frames = 16
        >>> num_sequences = 6
        >>> # Setup 5 sequences to norm
        >>> heatmaps = [np.empty(num_sequences) for _ in range(num_frames)]
        >>> heatmaps = np.array(heatmaps)
        >>> # Sequence 0 is all nan
        >>> heatmaps[:, 0] = np.nan
        >>> # Sequence 1 is random
        >>> heatmaps[:, 1] = np.random.rand(num_frames)
        >>> # Sequence 2 is Sequence1, but half of the data is nan
        >>> heatmaps[0:, 2] = heatmaps[:, 1]
        >>> heatmaps[0:num_frames // 2, 2] = np.nan
        >>> # Sequence 3 is all zero except for an impulse
        >>> heatmaps[0:, 3] = 0
        >>> heatmaps[num_frames // 2, 3] = 1
        >>> # Sequence 4 is a gaussian response
        >>> heatmaps[0:, 4] = kwimage.gaussian_patch(shape=(1, num_frames))[0]
        >>> # Sequence 5 is a a gaussian response 1 / 4 nans
        >>> heatmaps[0:, 5] = kwimage.gaussian_patch(shape=(1, num_frames))[0]
        >>> heatmaps[0:num_frames // 4, 5] = np.nan
        >>> norm_ord = 1
        >>> x = _norm(heatmaps, norm_ord)
        >>> y = np.linalg.norm(heatmaps, ord=norm_ord, axis=0)
        >>> print('heatmaps = {}'.format(ub.urepr(heatmaps, nl=1, precision=2)))
        >>> print(x)
        >>> print(y)
        >>> # xdoctest: +REQUIRES(--show)
        >>> # Visualize how this works for random signals
        >>> import kwplot
        >>> sns = kwplot.sns
        >>> kwplot.plt.ion()
        >>> # kwplot.close_figures()
        >>> # Add in the original signals
        >>> rows = []
        >>> for c in range(num_sequences):
        >>>     for x in range(num_frames):
        >>>         rows.append(
        >>>             {'x': x, 'col': c, 'ord': 'raw-signal', 'value': heatmaps[x, c]})
        >>> #
        >>> import pandas as pd
        >>> import scipy.stats
        >>> for norm_ord in [1, 2, 3, float('inf')]:
        >>>     v1 = _norm(heatmaps, norm_ord)
        >>>     v2 = scipy.stats.pmean(heatmaps, p=norm_ord, axis=0, nan_policy='omit')
        >>>     print(f'norm_ord={norm_ord}')
        >>>     print(f'v1={v1}')
        >>>     print(f'v2={v2}')
        >>>     for c in range(num_sequences):
        >>>         for x in range(num_frames):
        >>>             rows.append({'x': x, 'col': c, 'ord': norm_ord, 'value': v1[c]})
        >>>     ...
        >>> df = pd.DataFrame(rows)
        >>> pnum_ = kwplot.PlotNums(nSubplots=num_sequences)
        >>> for c in range(num_sequences):
        >>>     kwplot.figure(fnum=1, pnum=pnum_())
        >>>     subdata = df[df['col'] == c]
        >>>     sns.lineplot(data=subdata, x='x', y='value', hue='ord')
    """
    import numpy as np
    heatmaps = np.array(heatmaps)
    if norm_ord == 0:
        import scipy.stats
        probs = scipy.stats.pmean(heatmaps, p=norm_ord, axis=0, nan_policy='omit')
        probs = np.nan_to_num(probs)
    elif norm_ord == np.inf:
        probs = np.nanmax(heatmaps, axis=0)
    else:
        # The np.linalg.norm part
        probs = np.power(np.nansum(np.power(heatmaps, norm_ord), axis=0),
                         1. / norm_ord)
        if norm_ord > 0:
            n_nonzero = np.count_nonzero(~np.isnan(heatmaps), axis=0)
            # Force the denominator to be positive.
            n_nonzero[n_nonzero == 0] = 1
            probs /= np.power(n_nonzero, 1. / norm_ord)
    return probs


# give all these the same signature so they can be swapped out


def binary(heatmaps, norm_ord, morph_kernel, thresh, viz_dpath=None):
    import kwimage
    probs = _norm(heatmaps, norm_ord)

    hard_probs = kwimage.morphology(probs > thresh, 'dilate', morph_kernel)

    return hard_probs.astype('float')


def rescaled_binary(heatmaps, norm_ord, morph_kernel, thresh, upper_quantile=0.999, viz_dpath=None):
    import kwimage
    import kwarray
    import numpy as np
    probs = _norm(heatmaps, norm_ord)
    probs = kwarray.normalize(probs, min_val=0, max_val=np.quantile(probs, upper_quantile))

    hard_probs = kwimage.morphology(probs > thresh, 'dilate', morph_kernel)

    return hard_probs.astype('float')


def probs(heatmaps, norm_ord, morph_kernel, thresh, viz_dpath=None):
    import kwimage
    probs = _norm(heatmaps, norm_ord)

    hard_probs = kwimage.morphology(probs > thresh, 'dilate', morph_kernel)
    modulated_probs = probs * hard_probs

    if viz_dpath is not None:
        kwimage.imwrite(viz_dpath / 'probs_raw.png', kwimage.ensure_uint255(probs))
        kwimage.imwrite(viz_dpath / 'probs_hard.png', kwimage.ensure_uint255(hard_probs))
        kwimage.imwrite(viz_dpath / 'probs_modulated.png', kwimage.ensure_uint255(modulated_probs))

    return modulated_probs


def rescaled_probs(heatmaps, norm_ord, morph_kernel, thresh, upper_quantile=0.999, viz_dpath=None):
    import kwimage
    import kwarray
    import numpy as np
    probs = _norm(heatmaps, norm_ord)
    probs = kwarray.normalize(probs, min_val=0, max_val=np.quantile(probs, upper_quantile))

    hard_probs = kwimage.morphology(probs > thresh, 'dilate', morph_kernel)
    modulated_probs = probs * hard_probs

    return modulated_probs


def mean_normalized(heatmaps, norm_ord=1, morph_kernel=1, thresh=None, viz_dpath=None):
    """
    Normalize average_heatmap by applying a scaling based on max(heatmaps) and
    max(average_heatmap)
    """
    import numpy as np
    import kwimage
    average = _norm(heatmaps, norm_ord)

    scale_factor = np.max(heatmaps) / (np.max(average) + 1e-9)
    print('max heatmaps', np.max(heatmaps))
    print('max average', np.max(average))

    # average *= scale_factor
    average = 0.75 * average * scale_factor
    print('scale_factor', scale_factor)
    print('After scaling, max:', np.max(average))

    average = kwimage.morphology(average, 'dilate', morph_kernel)

    return average


def frequency_weighted_mean(heatmaps, thresh, norm_ord=0, morph_kernel=3, viz_dpath=None):
    """
    Convert a list of heatmaps to an aggregated score, averaging is computed
    based on samples for every pixel
    """
    import kwimage
    import numpy as np
    heatmaps = np.array(heatmaps)

    masks = 1 * (heatmaps > thresh)
    pixel_wise_samples = masks.sum(0) + 1e-9
    print('pixel_wise_samples', pixel_wise_samples)

    # compute sum
    aggregated_probs = _norm(masks * heatmaps, norm_ord)

    # divide by number of samples for every pixel
    aggregated_probs /= pixel_wise_samples

    aggregated_probs = kwimage.morphology(aggregated_probs, 'dilate',
                                          morph_kernel)

    return aggregated_probs


AGG_FN_REGISTRY = {
    'frequency_weighted_mean': frequency_weighted_mean,
    'mean_normalized': mean_normalized,
    'rescaled_probs': rescaled_probs,
    'probs': probs,
    'rescaled_binary': rescaled_binary,
    'binary': binary,
}



#!/usr/bin/env python3
"""
TODO: rectify with run_teamfeat_landcover.py
"""
import ubelt as ub
import scriptconfig as scfg
from geowatch.cli.smartflow_ingress import smartflow_ingress
from geowatch.cli.smartflow_egress import smartflow_egress


class TeamFeatLandcover(scfg.DataConfig):
    """
    Run DZYNE landcover feature computation as baseline framework component
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
        '''
        Path to the STAC items this step can use as inputs.
        This is usually an S3 Path.
        '''), alias=['input_stac_path'])

    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
        '''
        Path to input T&E Baseline Framework Region definition JSON
        '''))

    output_path = scfg.Value(None, type=str, position=3, required=True, help=ub.paragraph(
        '''
        Path to the STAC items that register the outputs of this stage.
        This is usually an S3 Path.
        '''), alias=['output_stac_path'])

    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        AWS Profile to use for AWS S3 CLI commands
        '''))

    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='Run AWS CLI commands with --dryrun flag')

    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
        '''
        S3 Output directory for STAC item / asset egress
        '''))

    newline = scfg.Value(False, isflag=True, short_alias=['n'], help=ub.paragraph(
        '''
        Output as simple newline separated STAC items
        '''))

    expt_dvc_dpath = scfg.Value('/root/data/smart_expt_dvc', help='location of the experiment DVC repo')


def main():
    config = TeamFeatLandcover.cli(strict=True)

    print('config = {}'.format(ub.urepr(config, nl=1, align=':')))
    from geowatch.utils.util_framework import download_region

    from geowatch.utils.util_framework import NodeStateDebugger
    node_state = NodeStateDebugger()
    node_state.print_environment()
    node_state.print_local_invocation(config)

    # 1. Ingress data
    print("* Running baseline framework kwcoco ingress *")

    # TODO: these input output bucket names need to be configurable so they can
    # be run at BAS or at ACSC time and composed at the DAG level.
    ingress_dir = ub.Path('/tmp/ingress')
    ingressed_assets = smartflow_ingress(
        config.input_path,
        [
            # Pull the current teamfeature-enriched dataset to modify
            'enriched_acsc_kwcoco_file',
            'enriched_acsc_kwcoco_teamfeats',
            'enriched_acsc_kwcoco_rawbands',
        ],
        ingress_dir,
        config.aws_profile,
        config.dryrun)

    print('ingressed_assets = {}'.format(ub.urepr(ingressed_assets, nl=1)))

    # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = ub.Path('/tmp/region.json')
    download_region(
        input_region_path=config.input_region_path,
        output_region_path=local_region_path,
        aws_profile=config.aws_profile,
        strip_nonregions=True,
    )

    ingress_dir_contents1 = list(ingress_dir.ls())
    print('ingress_dir_contents1 = {}'.format(ub.urepr(ingress_dir_contents1, nl=1)))

    input_kwcoco_fpath = ingressed_assets['enriched_acsc_kwcoco_file']

    # TOOD: better passing of configs

    # Use the existing prepare teamfeat script to get the features invocation.
    # This has a specific output pattern that we hard code here.
    from geowatch.cli.queue_cli import prepare_teamfeats
    base_fpath = ub.Path(input_kwcoco_fpath)
    # watch_coco_stats.main(cmdline=0, src=base_fpath)
    # coco_stats._CLI.main(cmdline=0, src=[base_fpath])

    node_state.print_current_state(ingress_dir)

    # ub.cmd(f'kwcoco validate {base_fpath}', verbose=3)
    ub.cmd(f'kwcoco stats {base_fpath}', verbose=3)
    ub.cmd(f'geowatch stats {base_fpath}', verbose=3)

    teamfeat_info = prepare_teamfeats.main(
        cmdline=0,
        with_wv_landcover=1,
        with_s2_landcover=1,
        num_wv_landcover_hidden=0,
        num_s2_landcover_hidden=0,
        expt_dvc_dpath=config.expt_dvc_dpath,
        base_fpath=base_fpath,
        assets_dname='_teamfeats',
        run=1,
        backend='serial',
    )
    final_output_paths = teamfeat_info['final_output_paths']
    assert len(final_output_paths) == 1
    full_output_kwcoco_fpath = final_output_paths[0]

    ingress_dir_contents2 = list(ingress_dir.ls())
    print('ingress_dir_contents2 = {}'.format(ub.urepr(ingress_dir_contents2, nl=1)))

    # Reroot kwcoco files to make downloaded results easier to work with
    ub.cmd(['kwcoco', 'reroot', f'--src={full_output_kwcoco_fpath}', '--inplace=1', '--absolute=0'])

    ub.cmd(f'kwcoco stats {full_output_kwcoco_fpath}', verbose=3)
    ub.cmd(f'geowatch stats {full_output_kwcoco_fpath}', verbose=3)

    print("* Egressing KWCOCO dataset and associated STAC item *")

    # This is the location that COLD features will be written to.
    teamfeat_dpath = (ingress_dir / '_teamfeats').ensuredir()
    (teamfeat_dpath / 'dummy').touch()
    ingressed_assets['enriched_acsc_kwcoco_teamfeats'] = teamfeat_dpath
    # This is the kwcoco file with the all teamfeature outputs (i.e. previous
    # team features + this one)
    ingressed_assets['enriched_acsc_kwcoco_file'] = full_output_kwcoco_fpath

    node_state.print_current_state(ingress_dir)

    smartflow_egress(ingressed_assets,
                     local_region_path,
                     config.output_path,
                     config.outbucket,
                     aws_profile=config.aws_profile,
                     dryrun=config.dryrun,
                     newline=config.newline)

    print('Finish run_teamfeat_cold')


if __name__ == "__main__":
    main()



"""
Baseline Example:

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
    DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)
    MAE_MODEL_FPATH="$DVC_EXPT_DPATH/models/wu/wu_mae_2023_04_21/Drop6-epoch=01-val_loss=0.20.ckpt"
    KWCOCO_BUNDLE_DPATH=$DVC_DATA_DPATH/Drop7-MedianNoWinter10GSD

    python -m geowatch.utils.simple_dvc request "$MAE_MODEL_FPATH"

    python -m geowatch.tasks.mae.predict \
        --device="cuda:0" \
        --mae_ckpt_path="$MAE_MODEL_FPATH" \
        --input_kwcoco="$KWCOCO_BUNDLE_DPATH/imganns-KR_R001.kwcoco.zip"\
        --output_kwcoco="$KWCOCO_BUNDLE_DPATH/imganns-KR_R001-testmae2.kwcoco.zip" \
        --window_space_scale=1.0 \
        --workers=8 \
        --assets_dname=teamfeats2 \
        --io_workers=8

    # After your model predicts the outputs, you should be able to use the
    # geowatch visualize tool to inspect your features.
    python -m geowatch visualize "$KWCOCO_BUNDLE_DPATH/imganns-KR_R001-testmae.kwcoco.zip" \
        --channels "red|green|blue,mae.8:11,mae.14:17" --stack=only --workers=avail --animate=True \
        --draw_anns=False

    # Batch computing

    export CUDA_VISIBLE_DEVICES="1"
    DVC_DATA_DPATH=$(geowatch_dvc --tags=phase2_data --hardware="hdd")
    DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware='auto')
    BUNDLE_DPATH=$DVC_DATA_DPATH/Drop7-MedianNoWinter10GSD
    python -m geowatch.cli.queue_cli.prepare_teamfeats \
        --base_fpath "$BUNDLE_DPATH"/imganns-*[0-9].kwcoco.zip \
        --expt_dvc_dpath="$DVC_EXPT_DPATH" \
        --with_mae=1 \
        --skip_existing=1 \
        --assets_dname=teamfeats \
        --gres=0,1 --tmux_workers=8 --backend=tmux --run=1
"""
import ubelt as ub
from kwutil import util_parallel
import scriptconfig as scfg
import albumentations as A
import kwimage
import kwcoco
import ndsampler
import torch
import torch.nn as nn
from pytorch_lightning import LightningModule
from torch.nn import L1Loss as MSE
from torch.utils.data import DataLoader, Dataset
from einops import rearrange
from einops.layers.torch import Rearrange
import numpy as np
from geowatch.tasks.fusion.predict import CocoStitchingManager


class MAEPredictConfig(scfg.DataConfig):
    """
    Configuration for WashU MAE models
    """
    device = scfg.Value('cuda:0', type=str)
    mae_ckpt_path = scfg.Value(None, type=str)
    batch_size = scfg.Value(1, type=int)
    workers = scfg.Value(4, help=ub.paragraph(
            '''
            number of background data loading workers
            '''), alias=['num_workers'])
    io_workers = scfg.Value(8, help=ub.paragraph(
            '''
            number of background data writing workers
            '''), alias=['write_workers'])

    window_resolution = scfg.Value(1.0, help='The window GSD to build the grid at', alias=['window_space_scale'])

    sensor = scfg.Value(['S2', 'L8'], nargs='+')
    bands = scfg.Value(['shared'], type=str, help=ub.paragraph(
            '''
            Choose bands on which to train. Can specify 'all' for all
            bands from given sensor, or 'share' to use common bands when
            using both S2 and L8 sensors
            '''), nargs='+')
    patch_overlap = scfg.Value(0.25, type=float)
    input_kwcoco = scfg.Value(None, type=str, required=True, help=ub.paragraph(
            '''
            Path to kwcoco dataset with images to generate feature for
            '''))
    output_kwcoco = scfg.Value(None, type=str, required=True, help=ub.paragraph(
            '''
            Path to write an output kwcoco file. Output file will be a
            copy of input_kwcoco with addition feature fields generated
            by predict.py rerooted to point to the original data.
            '''))

    assets_dname = scfg.Value('_assets', help=ub.paragraph(
        '''
        The name of the top-level directory to write new assets.
        '''))


class WatchDataset(Dataset):
    """

    Example:
        >>> # xdoctest: +REQUIRES(env:DVC_DPATH)
        >>> from geowatch.tasks.mae.predict import *  # NOQA
        >>> import geowatch
        >>> import kwcoco
        >>> import ubelt as ub
        >>> dvc_dpath = geowatch.find_dvc_dpath(tags='drop7_data', hardware='auto')
        >>> coco_fpath = dvc_dpath / 'Drop7-Cropped2GSD/BR_R002/BR_R002.kwcoco.zip'
        >>> self = WatchDataset(coco_fpath)
        >>> for idx in ub.ProgIter(range(len(self))):
        >>>     images, item = self[idx]

    """

    S2_l2a_channel_names = [
        'B02.tif', 'B01.tif', 'B03.tif', 'B04.tif', 'B05.tif', 'B06.tif', 'B07.tif', 'B08.tif', 'B09.tif', 'B11.tif', 'B12.tif', 'B8A.tif'
    ]
    S2_channel_names = [
        'coastal', 'blue', 'green', 'red', 'B05', 'B06', 'B07', 'nir', 'B09', 'cirrus', 'swir16', 'swir22', 'B8A'
    ]
    L8_channel_names = [
        'coastal', 'lwir11', 'lwir12', 'blue', 'green', 'red', 'nir', 'swir16', 'swir22', 'pan', 'cirrus'
    ]

    def __init__(self, coco_dset, sensor=['S2'], bands=['shared'],
                 segmentation=False, patch_size=224, mask_patch_size=16, num_images=2,
                 mode='train', patch_overlap=.25, bas=True, rng=None, mask_pct=.5, mask_time_width=2,
                 temporal_mode='cat', window_space_scale=1.0):
        super().__init__()
        if not isinstance(bands, list):
            bands = [bands]
        if not isinstance(sensor, list):
            sensor = [sensor]
        assert (temporal_mode in ['cat', 'stack'])

        # initialize dataset
        print('load dataset')
        self.coco_dset: kwcoco.CocoDataset = kwcoco.CocoDataset.coerce(coco_dset)

        print('filter dataset')
        # Filter out worldview images (better to use subset than remove)
        images: kwcoco.coco_objects1d.Images = self.coco_dset.images()
        flags = [s in sensor for s in images.lookup('sensor_coarse')]
        valid_image_ids : list[int] = list(images.compress(flags))
        self.coco_dset = self.coco_dset.subset(valid_image_ids)

        self.images : kwcoco.coco_objects1d.Images = self.coco_dset.images()
        self.sampler = ndsampler.CocoSampler(self.coco_dset)

        window_dims = [patch_size, patch_size]
        time_dims = num_images

        NEW_GRID = 1
        if NEW_GRID:
            print('make grid')
            from geowatch.tasks.fusion.datamodules.kwcoco_video_data import sample_video_spacetime_targets
            sample_grid = sample_video_spacetime_targets(
                self.coco_dset, window_dims=window_dims,
                time_dims=time_dims, window_overlap=patch_overlap,
                time_sampling='hardish3', time_span='1y',
                use_annot_info=False,
                keepbound=True,
                use_centered_positives=False,
                window_space_scale=window_space_scale
            )

            samples = sample_grid['targets']
            for tr in samples:
                tr['vidid'] = tr['video_id']  # hack

            WORKAROUND_NON_UNIQUE_IMAGE_IDS = 1
            if WORKAROUND_NON_UNIQUE_IMAGE_IDS:
                # FIXME: there is an issue in sample_video_spacetime_targets.
                # It should not be producing duplicate image ids. Workaround it
                # for now, but fix it for real later.
                workaround_samples = []
                for tr in samples:
                    unique_gids = list(ub.unique(tr['gids']))
                    if len(unique_gids) == time_dims:
                        tr['gids'] = unique_gids
                        workaround_samples.append(tr)
                samples = workaround_samples

            print('made grid')
        else:
            grid = self.sampler.new_sample_grid(**{
                'task': 'video_detection',
                'window_dims': [num_images, patch_size, patch_size],
                'window_overlap': patch_overlap,
            })
            if segmentation:
                samples = grid['positives']
            else:
                samples = grid['positives'] + grid['negatives']

        # vidid_to_patches = ub.group_items(samples, key=lambda x: x['vidid'])
        # self.vidid_to_patches = vidid_to_patches
        print('build patches')
        grouped = ub.group_items(
                samples,
                lambda x: tuple(
                    [x['vidid']] + [gid for gid in x['gids']]
                )
                )
        grouped = ub.sorted_keys(grouped)
        self.patches : list[dict] = list(ub.flatten(grouped.values()))
        self.bands = []
        # no channels selected
        if len(bands) < 1:
            raise ValueError(f'bands must be specified. Options are {", ".join(bands)}, or all')
        # all channels selected
        elif len(bands) == 1:
            if bands[0].lower() == 'all':
                self.bands = bands
            elif bands[0].lower() == 'shared':
                self.bands = ['red', 'green', 'blue', 'nir', 'swir16', 'swir22']
            elif bands[0] == 'r|g|b':
                self.bands.append('r|g|b')

        self.num_channels = len(self.bands)
        self.bands = "|".join(self.bands)

        # define augmentations
        print('build augs')
        additional_targets = dict()
        self.num_images = num_images

        for i in range(self.num_images):
            additional_targets['image{}'.format(1 + i)] = 'image'
            additional_targets['seg{}'.format(i + 1)] = 'mask'

        self.transforms = A.NoOp()

        self.mode = mode
        self.segmentation = segmentation
        self.patch_size = patch_size
        self.bas = bas
        if self.bas:
            self.positive_indices = [0, 1, 3]
            self.ignore_indices = [2, 6]
        else:
            self.positive_indices = [0, 1, 2, 3]
            self.ignore_indices = [6]
        print('finished dataset init')

        self.temporal_mode = temporal_mode

    def __len__(self):
        return len(self.patches)

    def __getitem__(self, idx):
        #if idx > 500: raise IndexError
        tr : dict = self.patches[idx]
        tr['channels'] = self.bands
        tr = self.update_target_properties(tr)
        # vidid = tr['vidid']
        gids : list[int] = tr['gids']

        sample = self.sampler.load_sample(tr, nodata='float')
        images : np.ndarray = sample['im']
        images = images.astype(np.float32)
        std = np.nanstd(images)
        mean = np.nanmean(images)
        if std != 0:
            images = np.nan_to_num((images - mean) / std)
        else:
            images = np.zeros_like(images)

        if self.temporal_mode == 'cat':
            images = torch.cat([torch.tensor(x) for x in images], dim=0).permute(2, 0, 1)
        else:
            images = torch.tensor(images).permute(0, 3, 1, 2)

        vidspace_box = kwimage.Box.from_slice(tr['space_slice'])

        scale_outspace_from_vidspace = tr['scale'] / 4  # Add it back
        outspace_box = vidspace_box.scale(scale_outspace_from_vidspace).quantize().astype(np.int32)
        item = dict()
        im1_id = gids[0]
        img_obj1 : dict = self.coco_dset.index.imgs[im1_id]
        video_obj = self.coco_dset.index.videos[img_obj1['video_id']]

        full_stitch_vidspace_box = kwimage.Box.coerce([0, 0, video_obj['width'], video_obj['height']], format='xywh')
        full_stitch_outspace_box = full_stitch_vidspace_box.scale(scale_outspace_from_vidspace).quantize().astype(np.int32)

        item['full_stitch_outspace_ltrb'] = torch.from_numpy(full_stitch_outspace_box.data)
        item['sample_outspace_ltrb'] = torch.from_numpy(outspace_box.data)
        item['scale_outspace_from_vid'] = scale_outspace_from_vidspace

        return images, item

    def update_target_properties(self, target):
        """
        Populate the target so it has the correct input scale and bands.
        """
        # Handle target scale
        from geowatch.tasks.fusion.datamodules import data_utils
        gids : list[int] = target['gids']
        im1_id = gids[0]
        img_obj1 : dict = self.coco_dset.index.imgs[im1_id]
        video_obj = self.coco_dset.index.videos[img_obj1['video_id']]
        vidspace_gsd = video_obj.get('target_gsd', None)
        resolved_input_scale = data_utils.resolve_scale_request(request=1.0, data_gsd=vidspace_gsd)
        target['scale'] = resolved_input_scale['scale']
        target['channels'] = self.bands
        target['_input_gsd'] = resolved_input_scale['gsd']
        target['_native_video_gsd'] = resolved_input_scale['data_gsd']
        return target


def pair(t):
    return t if isinstance(t, tuple) else (t, t)


class PreNorm(nn.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.norm = nn.LayerNorm(dim)
        self.fn = fn

    def forward(self, x, **kwargs):
        return self.fn(self.norm(x), **kwargs)


class FeedForward(nn.Module):
    def __init__(self, dim, hidden_dim, dropout=0.):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, hidden_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(hidden_dim, dim),
            nn.Dropout(dropout)
        )

    def forward(self, x):
        return self.net(x)


class Attention(nn.Module):
    def __init__(self, dim, heads=8, dim_head=64, dropout=0.):
        super().__init__()
        inner_dim = dim_head *  heads
        project_out = not (heads == 1 and dim_head == dim)

        self.heads = heads
        self.scale = dim_head ** -0.5

        self.attend = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)

        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)

        self.to_out = nn.Sequential(
            nn.Linear(inner_dim, dim),
            nn.Dropout(dropout)
        ) if project_out else nn.Identity()

    def forward(self, x):
        qkv = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h=self.heads), qkv)

        dots = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        attn = self.attend(dots)
        attn = self.dropout(attn)

        out = torch.matmul(attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')
        return self.to_out(out)


class Transformer(nn.Module):
    def __init__(self, dim, depth, heads, dim_head, mlp_dim, dropout=0.):
        super().__init__()
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(nn.ModuleList([
                PreNorm(dim, Attention(dim, heads=heads, dim_head=dim_head, dropout=dropout)),
                PreNorm(dim, FeedForward(dim, mlp_dim, dropout=dropout))
            ]))

    def forward(self, x):
        for attn, ff in self.layers:
            x = attn(x) + x
            x = ff(x) + x
        return x


class ViT(nn.Module):
    def __init__(self, *, image_size, image_patch_size, frames, frame_patch_size, dim, depth, heads, mlp_dim, channels=6, dim_head=64, dropout=0., emb_dropout=0.):
        super().__init__()
        image_height, image_width = pair(image_size)
        patch_height, patch_width = pair(image_patch_size)

        assert image_height % patch_height == 0 and image_width % patch_width == 0, 'Image dimensions must be divisible by the patch size.'
        assert frames % frame_patch_size == 0, 'Frames must be divisible by frame patch size'

        num_patches = (image_height // patch_height) * (image_width // patch_width) * (frames // frame_patch_size)
        patch_dim = channels * patch_height * patch_width * frame_patch_size

        self.to_patch_embedding = nn.Sequential(
            Rearrange('b (f pf) c (h p1) (w p2) -> b (f h w) (p1 p2 pf c)', p1=patch_height, p2=patch_width, pf=frame_patch_size),
            nn.Linear(patch_dim, dim),
        )

        self.pos_embedding = nn.Parameter(torch.randn(1, num_patches + 1, dim))

        self.dropout = nn.Dropout(emb_dropout)

        self.transformer = Transformer(dim, depth, heads, dim_head, mlp_dim, dropout)

    def forward(self, video):
        x = self.to_patch_embedding(video)
        b, n, _ = x.shape

        x += self.pos_embedding[:, :(n + 1)]
        x = self.dropout(x)

        x = self.transformer(x)

        return x


class MAE(nn.Module):
    def __init__(
        self,
        *,
        encoder,
        decoder_dim,
        masking_ratio=0.75,
        decoder_depth=8,
        decoder_heads=8,
        decoder_dim_head=64
    ):
        super().__init__()
        assert masking_ratio > 0 and masking_ratio < 1, 'masking ratio must be kept between 0 and 1'
        self.masking_ratio = masking_ratio

        # extract some hyperparameters and functions from encoder (vision transformer to be trained)

        self.encoder = encoder
        num_patches, encoder_dim = encoder.pos_embedding.shape[-2:]
        self.to_patch, self.patch_to_emb = encoder.to_patch_embedding[:2]
        pixel_values_per_patch = self.patch_to_emb.weight.shape[-1]

        # decoder parameters
        self.decoder_dim = decoder_dim
        self.enc_to_dec = nn.Linear(encoder_dim, decoder_dim) if encoder_dim != decoder_dim else nn.Identity()
        self.mask_token = nn.Parameter(torch.randn(decoder_dim))
        self.decoder = Transformer(dim=decoder_dim, depth=decoder_depth, heads=decoder_heads, dim_head=decoder_dim_head, mlp_dim=decoder_dim * 4)
        #self.decoder_pos_emb = nn.Embedding(num_patches, decoder_dim)
        self.decoder_pos_emb = nn.Parameter(torch.randn(num_patches, decoder_dim))
        self.to_pixels = nn.Linear(decoder_dim, pixel_values_per_patch)
        self.out = nn.Sigmoid()

    def forward(self, img):

        patches = self.to_patch(img)

        tokens = self.patch_to_emb(patches)
        tokens = tokens + self.encoder.pos_embedding[:, 1:(tokens.shape[1] + 1)]

        encoded_tokens = self.encoder.transformer(tokens)

        encoded_tokens = rearrange(encoded_tokens, 'b (f h w) d -> b f h w d', h=32, w=32)

        return encoded_tokens


class MaeCityscape(LightningModule):
    def __init__(self, dataset, **kwargs):
        super().__init__()
        self.vit = ViT(
            image_size=128,
            image_patch_size=4,
            frames=4,
            frame_patch_size=2,
            dim=16,
            depth=12,
            heads=12,
            mlp_dim=1024,
            dropout=0.1
        )
        self.model = MAE(
            encoder=self.vit,
            masking_ratio=0.90,
            decoder_dim=64,
            decoder_depth=6,
        )
        self.dataset = dataset
        self.batch_size = kwargs.get('batch_size', 4)
        self.num_workers = kwargs.get('num_workers', 16)
        self.lr = kwargs.get('lr', 0.02)
        self.acc = MSE()

    def forward(self, x):
        return self.model(x)

    def shared_step(self, batch, batch_idx):
        (x, masks), dates = batch

        pred, gt, viz, mi, ui = self(x)
        #gt = repeat(gt, 'b n d -> b (n n2) d', n2= 2)
        batch_range = torch.arange(x.shape[0], device=x.device)[:, None]
        loss = 0.999 * self.acc(pred[batch_range, mi], gt[batch_range, mi]) + 0.001 * self.acc(pred[batch_range, ui], gt[batch_range, ui])
        #loss = self.acc(pred, gt)
        return loss, viz, pred, gt


def sigmoid(a):
    return 1 / (1 + np.exp(-a))


class Predict():
    def __init__(self, args):

        self.device = torch.device(args.device)
        self.data_path = args.input_kwcoco
        self.dataset = WatchDataset(self.data_path, sensor=args.sensor, bands=args.bands,
                                    segmentation=False, patch_size=128, mask_patch_size=16, num_images=4,
                                    mode='train', mask_pct=0.5, patch_overlap=args.patch_overlap,
                                    temporal_mode='stack',
                                    mask_time_width=2, window_space_scale=args.window_resolution)

        print("Dataset load finished ...")

        self.model = MaeCityscape(self.dataset)

        self.model = self.model.load_from_checkpoint(args.mae_ckpt_path, dataset=self.dataset)
        print("Model load finished ...")

        self.dataloader = DataLoader(self.dataset,
                                     shuffle=False,
                                     batch_size=1,
                                     num_workers=args.workers,
                                     persistent_workers=False,
                                     pin_memory=True)

        print('copy dataset')
        self.output_dset = self.dataset.coco_dset.copy()

        print('reroot')
        self.output_dset.reroot(absolute=True)
        self.output_dset.fpath = args.output_kwcoco
        self.output_dset.reroot(absolute=False)

        self.save_channels = 'mae.0:16'
        self.output_kwcoco_path = ub.Path(self.output_dset.fpath)
        out_folder = self.output_kwcoco_path.parent
        self.output_feat_dpath = (out_folder / args.assets_dname).ensuredir()

        self.imwrite_kw = {
            'compress': 'DEFLATE',
            'backend': 'gdal',
            'blocksize': 128,
        }

        self.stitch_manager = CocoStitchingManager(
            result_dataset=self.output_dset,
            short_code='mae',
            chan_code=self.save_channels,
            stiching_space='video',
            prob_compress=self.imwrite_kw['compress'],
            quantize=True,
            assets_dname=args.assets_dname,
        )

        from geowatch.utils import process_context
        self.proc_context = process_context.ProcessContext(
            type='process',
            name='geowatch.tasks.mae.predict',
            config=dict(args),
        )

    def __call__(self):

        writer_queue = util_parallel.BlockingJobQueue(max_workers=4)
        self.stitch_manager.writer_queue = writer_queue

        self.proc_context.start()

        self.proc_context.add_disk_info(ub.Path(self.output_dset.fpath).parent)
        self.output_dset.dataset.setdefault('info', [])
        self.output_dset.dataset['info'].append(self.proc_context.obj)

        print('Evaluating and saving features')

        self.model.eval()
        self.model.to(self.device)
        num_batches = len(self.dataloader)
        preds = []
        with torch.no_grad():
            seen_images = set()
            prog = ub.ProgIter(enumerate(self.dataloader), total=num_batches, desc='Compute features', verbose=1)
            for batch_idx, batch in prog:
                x, item = batch
                x = x.to(self.device)
                x2 = rearrange(x, 'b (f pf) c h w -> b (pf f) c h w', pf=2)

                pred = self.model(x)
                pred2 = self.model(x2)
                preds = pred.cpu().detach().numpy()
                preds2 = pred2.cpu().detach().numpy()

                target = self.dataset.patches[batch_idx]

                new_complete_gids = target.get('new_complete_gids', [])

                for gid in new_complete_gids:
                    assert gid not in seen_images
                    seen_images.add(gid)
                    self.stitch_manager.submit_finalize_image(gid)

                gid1, gid2, gid3, gid4 = target['gids']

                sample_outspace_ltrb = kwimage.Box.coerce(item['sample_outspace_ltrb'].numpy(), format='ltrb')
                full_stitch_outspace_box = kwimage.Box.coerce(item['full_stitch_outspace_ltrb'].numpy(), format='ltrb')
                scale_outspace_from_vid = item['scale_outspace_from_vid'].numpy()[0]
                outspace_slice = sample_outspace_ltrb.to_slice()
                outspace_dsize = full_stitch_outspace_box.dsize

                feat1 = preds[:, 0, :, :, :].squeeze()
                feat2 = preds[:, 1, :, :, :].squeeze()
                feat3 = preds2[:, 0, :, :, :].squeeze()
                feat4 = preds2[:, 1, :, :, :].squeeze()

                self.stitch_manager.accumulate_image(
                    gid1, outspace_slice, feat1,
                    dsize=outspace_dsize,
                    scale=scale_outspace_from_vid)

                self.stitch_manager.accumulate_image(
                    gid2, outspace_slice, feat2,
                    dsize=outspace_dsize,
                    scale=scale_outspace_from_vid)

                self.stitch_manager.accumulate_image(
                    gid3, outspace_slice, feat3,
                    dsize=outspace_dsize,
                    scale=scale_outspace_from_vid)

                self.stitch_manager.accumulate_image(
                    gid4, outspace_slice, feat4,
                    dsize=outspace_dsize,
                    scale=scale_outspace_from_vid)

            print('Finalize already compelted jobs')
            writer_queue.wait_until_finished(desc='Finalize submitted jobs')

            for gid in ub.ProgIter(list(self.stitch_manager.image_stitchers.keys()), desc='submit loose write jobs'):
                if gid not in seen_images:
                    seen_images.add(gid)
                    self.stitch_manager.submit_finalize_image(gid)

            print('Finalize loose jobs')
            writer_queue.wait_until_finished()

        print('Finish process context')
        self.proc_context.add_device_info(self.device)
        self.proc_context.stop()

        print('Write to dset.fpath = {!r}'.format(self.output_dset.fpath))
        self.output_dset.dump(self.output_dset.fpath, newlines=True)
        print('Done')

        return


def main(cmdline=1, **kwargs):
    args = MAEPredictConfig.cli(cmdline=cmdline, data=kwargs)
    import rich
    rich.print('config = ' + ub.urepr(args))
    predict = Predict(args)
    predict()


if __name__ == '__main__':
    """
    SeeAlso:
        ../../cli/queue_cli/prepare_teamfeats.py

    CommandLine:
        python -m geowatch.tasks.template.predict --help

        python -m geowatch.tasks.mae.predict \
        --device="cuda:0"\
        --mae_ckpt_path="/storage1/fs1/jacobsn/Active/user_s.sastry/smart_watch/new_models/checkpoints/Drop6-epoch=01-val_loss=0.20.ckpt"\
        --input_kwcoco="$DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/data_train_I2L_split6.kwcoco.zip"\
        --output_kwcoco="$DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/mae_v1_train_split6.kwcoco.zip"\
        --window_space_scale=1.0 \
        --workers=8 \
        --io_workers=8

        python -m geowatch visualize $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/mae_v1_train_split6.kwcoco.zip \
        --channels "red|green|blue,mae.8:11,mae.14:17" --stack=only --workers=avail --animate=True \
        --draw_anns=False
    """
    main()



"""
Experimental feature where each training run will register itself in a global
registery. We can then provide a command to make it easy to check which
training runs were started on this machine and where they are located.


Usage
-----

To See

.. code:: bash

        python -m geowatch.tasks.fusion.fit_registery peek

"""
import ubelt as ub
import warnings
import scriptconfig as scfg


class Registery:
    """
    Attributes:
        fpath : path to the registery's shelf database.
    """
    __registery_fname__ = 'abstract.shelf'
    __appname__ = 'geowatch'
    __default__ = {}

    def __init__(self, fpath=None):
        if fpath is None:
            app_dpath = ub.Path.appdir(type='config', appname=self.__appname__)
            registry_dpath = (app_dpath / 'registry').ensuredir()
            fpath = registry_dpath / self.__registery_fname__

        self.fpath = fpath
        self._required_attributes = {k for k, v in self.__default__.items() if v is ub.NoParam}

    def pandas(self, **kwargs):
        import pandas as pd
        df = pd.DataFrame(self.query(**kwargs))
        return df

    def list(self, **kwargs):
        from rich import print
        print(self.pandas(**kwargs).to_string())

    def _open(self):
        import os
        import shelve
        shelf = shelve.open(os.fspath(self.fpath))
        return shelf

    def query(self, **kwargs):
        unexepcted = ub.udict(kwargs) - self.__default__
        if unexepcted:
            raise ValueError(
                'Unexpected query keywords: {}. Valid keywords are {}'.format(
                    ub.urepr(list(unexepcted.keys()), nl=0),
                    ub.urepr(list(self.__default__.keys()), nl=0),
                ))
        query = ub.udict({k: v for k, v in kwargs.items() if v is not None})

        results = []
        candidate_rows = self.read()
        for row in candidate_rows:
            if query:
                relevant = ub.udict(row).subdict(query, default=None)
                flag = relevant == query
            else:
                flag = True
            if flag:
                results.append(row)

        results = sorted(
            results,
            key=lambda r:
                r['priority']
                if r.get('priority', None) is not None else
                -float('inf'))[::-1]
        return results

    def _items(self):
        items = []
        with self._open() as shelf:
            try:
                items += list(shelf.items())
            except Exception as ex:
                warnings.warn('Unable to access shelf: {}'.format(ex))
        return items


class DictRegistery(Registery):

    def add(self, key, **kwargs):
        unknown = ub.udict(kwargs) - self.__default__
        if unknown:
            raise ValueError(f'Unknown kwargs={unknown}')

        for key in self._required_attributes:
            if key not in kwargs:
                raise ValueError(f'Must specify a {key}')

        row = ub.udict(self.__default__)
        row |= (kwargs & row)
        shelf = self._open()
        try:
            shelf[key] = row
        finally:
            shelf.close()

    def set(self, key, **kwargs):
        """
        Set an attribute of a row
        """
        if key is None:
            raise ValueError('Must specify a key')
        unknown = ub.udict(kwargs) - self.__default__
        if unknown:
            raise ValueError(f'Unknown kwargs={unknown}')
        row = ub.udict({'key': key}) | self.__default__
        row |= (kwargs & row)
        shelf = self._open()
        try:
            existing = shelf[key]
            row |= {existing[k] for k, v in row if v is None}
            shelf[key] = row
        finally:
            shelf.close()

    def remove(self, name):
        """
        Ignore:
            name = 'test'
            path = 'foo/bar'
            hardware = 'fake'
        """
        if name is None:
            raise ValueError('Must specify a name')
        shelf = self._open()
        try:
            shelf.pop(name)
        except Exception as ex:
            warnings.warn('Unable to access shelf: {}'.format(ex))
        finally:
            shelf.close()

    def read(self):
        registry_rows = []
        with self._open() as shelf:
            try:
                registry_rows += list(shelf.values())
            except Exception as ex:
                warnings.warn('Unable to access shelf: {}'.format(ex))
        return registry_rows


class ListRegistery(Registery):
    """
    Example:
        >>> # A minimal example that inherits from ListRegistery
        >>> from geowatch.tasks.fusion.fit_registery import *  # NOQA
        >>> class DemoListRegistery(ListRegistery):
        >>>     __registery_fname__ = 'geowatch_fit_registery.shelf'
        >>>     __appname__ = 'geowatch/demo'
        >>>     __default__ = ub.udict({
        >>>         'path'     : ub.NoParam,
        >>>         'time'     : None,
        >>>         'uuid'     : None,
        >>>         'config'   : None,
        >>>     })
        >>> #
        >>> self = DemoListRegistery()
        >>> self.fpath
        >>> print(self.fpath.exists())
        >>> self.append(path='foo', time='bar', uuid='blag')
        >>> self.append(path='foo', time='bar', uuid='blag')
        >>> list(self._open().items())
        >>> print(self.read())
        >>> print(self.pandas())
        >>> print(self.fpath.stat().st_size)
    """

    def append(self, **kwargs):
        unknown = ub.udict(kwargs) - self.__default__
        if unknown:
            raise ValueError(f'Unknown kwargs={unknown}')
        row = ub.udict(self.__default__)
        row |= (kwargs & row)
        shelf = self._open()
        try:
            if '__list__' not in shelf:
                shelf['__list__'] = []
            accum = shelf['__list__']
            accum.append(row)
            # Inefficient, can split across multiple entries giving each a max
            # size.
            shelf['__list__'] = accum
        finally:
            shelf.close()

    def read(self):
        registry_rows = []
        shelf = self._open()
        try:
            registry_rows += list(shelf['__list__'])
        except Exception as ex:
            warnings.warn('Unable to access shelf: {}'.format(ex))
        finally:
            shelf.close()
        print(f'registry_rows = {ub.urepr(registry_rows, nl=1)}')
        return registry_rows

    def pandas(self, **kwargs):
        import pandas as pd
        df = pd.DataFrame(self.query(**kwargs))
        return df


class NewDVCRegistery(DictRegistery):
    # TODO:
    __registery_fname__ = 'geowatch_new_dvc_registery.shelf'

    __default__ = ub.udict({
        'name': ub.NoParam,
        'path': ub.NoParam,
        'priority': None,
        'hardware': None,
        'tags': None,
    })


class FitRegistery(ListRegistery):
    """
    This is a shelf with a single list of rows that has a globally consistent
    path on each machine. The __default__ attribute specifies which keys each
    row will have. This is used to store information whenever training is run
    and stores the path to its output directory. This can be used to inspect
    which training runs artifacts might be available on this machine and also
    to help the user navigate to recent paths.

    Example:
        >>> from geowatch.tasks.fusion.fit_registery import *  # NOQA
        >>> class DemoFitRegistery(FitRegistery):
        >>>     __registery_fname__ = 'geowatch_demo_fit_registery.shelf'
        >>>     __appname__ = 'geowatch/demo'
        >>> #
        >>> self = DemoFitRegistery()
        >>> self.append(path=ub.Path('~').expand())
        >>> self.append(path=ub.Path('~').expand())
        >>> self.append(path=ub.Path('~').expand())
        >>> print(self.pandas())
        >>> print(self._items())

    Example:
        >>> # This is the real one, dont modify it
        >>> from geowatch.tasks.fusion.fit_registery import *  # NOQA
        >>> self = FitRegistery()
        >>> print(self.pandas())
        >>> print(self._items())
    """
    __registery_fname__ = 'geowatch_fit_registery.shelf'
    __default__ = ub.udict({
        'path'     : ub.NoParam,
        'uuid'     : None,
        'time'     : None,
        'user'     : None,
        'hostname' : None,
        'argv'     : None,
        'config'   : None,
    })

    def append(self, **kwargs):
        import os
        import uuid
        import platform
        path = kwargs.get('path', None)
        if path is not None:
            kwargs['path'] = os.fspath(path)

        path = kwargs.get('uuid', None)
        if path is None:
            kwargs['uuid'] = str(uuid.uuid4())

        time = kwargs.get('time', None)
        if time is None:
            kwargs['time'] = ub.timestamp()

        user = kwargs.get('user', None)
        if user is None:
            kwargs['user'] = ub.Path.home().name

        hostname = kwargs.get('hostname', None)
        if hostname is None:
            kwargs['hostname'] = platform.node()

        super().append(**kwargs)

    def peek(self):
        items = self.read()
        if not items:
            return None
        else:
            return items[-1]


class FitRegisteryCLI(scfg.ModalCLI):
    """
    Command line helper
    """
    __command__ = 'registery'

    class Append(scfg.DataConfig):
        """
        Register a new path (or overwrite / update an existing one)
        """
        __command__ = 'add'
        path = scfg.Value(None, help='path to register', position=1)

        @classmethod
        def main(cls, cmdline=1, **kwargs):
            config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
            registry = FitRegistery()
            registry.append(**config)

    class List(scfg.DataConfig):
        """
        List registered paths
        """
        __command__ = 'list'

        @classmethod
        def main(cls, cmdline=1, **kwargs):
            config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
            config = dict(config)
            registry = FitRegistery()
            registry.list(**config)

    class Peek(scfg.DataConfig):
        """
        Search for a path registered via ``sdvc registry add``
        """
        __command__ = 'peek'

        @classmethod
        def main(cls, cmdline=1, **kwargs):
            cls.cli(cmdline=cmdline, data=kwargs, strict=True)
            registry = FitRegistery()
            row = registry.peek()
            print(f'row = {ub.urepr(row, nl=1)}')


if __name__ == '__main__':
    """
    CommandLine:
        python -m geowatch.tasks.fusion.fit_registery peek
    """
    FitRegisteryCLI.main()
